----------------------------------------------------------------------------------
----------------------------------------------------------------------------------
-- Filename     NVMeHS_Test.vhd
-- Title        512-bit Axi-Master Write Controller
--
-- Company      Design Gateway Co., Ltd.
-- Project      SATA-IP
-- PJ No.       
-- Syntax       VHDL
-- Note         

-- Version      0.01
-- Author       B.Attapon
-- Date         2018/02/21
-- Remark       New Creation
----------------------------------------------------------------------------------
----------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.ALL;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;

Entity NVMeHS_Test Is
    Port 
	(	
		RstB				: in	std_logic;							-- Reset (Active low)
		Clk					: in	std_logic;							-- Clock

		-- Command I/F
		AxiWrReq			: in	std_logic;							-- Transfer Start pulse
		AxiWrLen			: in	std_logic_vector( 47 downto 0 ); 	-- Length in sector unit
		AxiWrBusy			: out	std_logic;							-- Transfer Busy status
		
		-- Data Control
		BuffWrCnt			: in	std_logic_vector( 30 downto 9 );	-- Buffer write counter
		CurBuffWrAdValid	: out	std_logic;							-- CurBuffWrAddr Valid pulse
		CurBuffWrAddr		: out	std_logic_vector( 30 downto 9 );	-- Current write buffer address
		
		-- Fifo
		AxiFfRdEn			: out	std_logic;
		AxiFfRdData			: in	std_logic_vector( 511 downto 0 );
		AxiFfRdCnt			: in	std_logic_vector( 15 downto 0 );
		
		-- AXI interface
		AXIAwAddr			: out	std_logic_vector( 30 downto 0 );
		AXIAwLen			: out	std_logic_vector( 7 downto 0 );
		AXIAwSize			: out	std_logic_vector( 2 downto 0 );
		AXIAwBurst			: out	std_logic_vector( 1 downto 0 );
		AXIAwValid			: out	std_logic;
		AXIAwReady			: in	std_logic;

		AXIwData			: out	std_logic_vector( 511 downto 0 );
		AXIwStrb			: out	std_logic_vector( 63 downto 0 );
		AXIwLast			: out	std_logic;
		AXIwValid			: out	std_logic;
		AXIwReady			: in	std_logic;

		AXIBReady			: out	std_logic;
		AXIBValid			: in	std_logic
	);							
End Entity NVMeHS_Test;

Architecture rtl Of NVMeHS_Test Is

----------------------------------------------------------------------------------
-- Signal declaration
----------------------------------------------------------------------------------
	
	type	AXIwStateType is
						(
							stIdle		,
							stChkWrRdy	,
							stGenWrReq	,
							stWrTrans	,
							stWtResp
						);
	signal	rState	: AXIwStateType;
	
	signal	rAxiWrBusy			: std_logic;						-- Indicates that AxiWrMt is woking when rAxiWrBusy='1'
	signal	rTrnLenCnt			: std_logic_vector( 47 downto 0 );	-- Remaining Write transfer length in sector unit
	signal	rAXIAwAddr			: std_logic_vector( 30 downto 9 );
	signal	rBurst8En			: std_logic;						-- Look ahead control signal for 8-sector transfer
	signal	rBurst4En			: std_logic;						-- Look ahead control signal for 4-sector transfer
	signal	rBurst2En			: std_logic;						-- Look ahead control signal for 2-sector transfer
	signal	rAXIAwLen			: std_logic_vector( 6 downto 0 );	-- unused rAXIAwLen(6)
	signal	rBurstLen			: std_logic_vector( 3 downto 0 );	-- BurstLen in sector unit
	signal	rAXIAwValid			: std_logic;
	signal	rBurstCnt			: std_logic_vector( 5 downto 0 );	-- decrement counter for indicate Number of Data that read form FIFO
	signal	rAXIwLast			: std_logic_vector( 1 downto 0 );
	signal	rAXIwValid			: std_logic;
	signal	rAxiFfRdDataLat		: std_logic_vector( 511 downto 0 );	-- Latch AxiFfRdData for ensure to met timing constrain
	signal	r2beatRdEn			: std_logic_vector( 1 downto 0 );	-- Indicate 2 step read data fifo before transfer
	signal	wAxiFfRdEn			: std_logic;						-- Read Fifo enable
	signal	rCurBuffWrAdValid	: std_logic;						-- Valid Pulse for indicate CurBuffWrAddr has changed
	signal	rCurBuffWrAddr		: std_logic_vector( 30 downto 9 );	-- Current DDR Write byte pointer
	signal	rAXIBValid			: std_logic;

Begin

----------------------------------------------------------------------------------
-- Output Assignment
----------------------------------------------------------------------------------

	-- AXI interface
	AXIAwAddr			<= rAXIAwAddr(30 downto 9)&('0'&x"00");	-- Always align 512-byte unit (1 burst size)
	AXIAwLen			<= "00"&rAXIAwLen(5 downto 0);			-- 8/16/32 beat
	AXIAwSize       	<= "110";								-- Fix write data port size to 512-bit
	AXIAwBurst			<= "01";								-- Auto increment address mode
	AXIAwValid			<= rAXIAwValid;

	AXIwData			<= rAxiFfRdDataLat;
	AXIwStrb			<= x"FFFF_FFFF_FFFF_FFFF";
	AXIwLast			<= rAXIwLast(1);
	AXIwValid			<= rAXIwValid;

	AXIBReady			<= '1';
	
	AxiWrBusy			<= rAxiWrBusy;
	CurBuffWrAdValid	<= rCurBuffWrAdValid;
	CurBuffWrAddr		<= rCurBuffWrAddr;
	
	AxiFfRdEn			<= wAxiFfRdEn;
	
----------------------------------------------------------------------------------
-- DFF
----------------------------------------------------------------------------------
	
	-- Axi Write Interface
	u_rState : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			if ( RstB='0' ) then
				rState		<= stIdle;
			else
				case ( rState ) is				
					-- Wait start pulse
					when stIdle		=>
						if ( AxiWrReq='1' ) then
							rState	<= stChkWrRdy;
						else
							rState	<= stIdle;
						end if;
					
					when stChkWrRdy	=>
						-- At least 1 sector available in FIFO
						-- More than 1 sector space available in Buffer
						if ( AxiFfRdCnt(15 downto 3)/=0 and BuffWrCnt(30 downto 10)/=('1'&x"FFFFF") ) then
							rState	<= stGenWrReq;
						else
							rState	<= stChkWrRdy;
						end if;

					when stGenWrReq	=>
						-- Bus receive command complete
                        if ( AXIAwReady='1' and rAXIAwValid='1' ) then
                            rState  <= stWrTrans;
                        else
                            rState  <= stGenWrReq;
                        end if;
						
					when stWrTrans	=>
						-- End of burst
						if ( AXIwReady='1' and rAXIwLast(1)='1' ) then
							rState	<= stWtResp;
						else
							rState	<= stWrTrans;
						end if;
					
					when stWtResp	=>
						-- Receive response from bus
                        if ( rAXIBValid='1' ) then
							if ( rTrnLenCnt(47 downto 0)=0 ) then
								rState	<= stIdle;
							else
								rState	<= stChkWrRdy;
							end if;
                        else
                            rState  <= stWtResp;
                        end if;
				end case;
			end if;
		end if;
	End Process u_rState;
	
	-- Generate Busy status
	u_rAxiWrBusy : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			if ( RstB='0' ) then
				rAxiWrBusy	<= '0';
			else
				if ( rState=stIdle ) then
					rAxiWrBusy	<= '0';
				else
					rAxiWrBusy	<= '1';
				end if;
			end if;
		end if;
	End Process u_rAxiWrBusy;
	
	-- Calculate Remaining Write transfer length and AXIAwAddr
	u_rTrnLenCnt : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			-- Load command length/Reset buffer address
			if ( rState=stIdle and AxiWrReq='1' ) then
				rTrnLenCnt(47 downto 0)		<= AxiWrLen(47 downto 0);
				rAXIAwAddr(30 downto 9)		<= (others=>'0');
			-- Write request received
			elsif ( rState=stGenWrReq and AXIAwReady='1' and rAXIAwValid='1' ) then
				-- Decrement remaining write transfer length in size of last BurstLen 
				rTrnLenCnt(47 downto 0)		<= rTrnLenCnt - (x"000_0000_0000"&rBurstLen(3 downto 0));
				-- Increment to next write address
				rAXIAwAddr(30 downto 9)		<= rAXIAwAddr(30 downto 9) + ("00"&x"0000"&rBurstLen(3 downto 0));
			else
				rTrnLenCnt(47 downto 0)		<= rTrnLenCnt(47 downto 0);
				rAXIAwAddr(30 downto 9)		<= rAXIAwAddr(30 downto 9);
			end if;
		end if;
	End Process u_rTrnLenCnt;
	
	-- Calculate look ahead control signal for select Size of Burst Transfer
	-- Generate rBurst8/4/2En at stWrTrans
	-- Using at stChkWrRdy
	u_rBurstEn : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			-- Enable Write 8 Sector
			-- 8 sector remain transfer size
			-- 8 sector alignment address
			if ( rTrnLenCnt(47 downto 3)/=0 and 
				 rAXIAwAddr(11 downto 9)="000" ) then
				rBurst8En	<= '1';
			else
				rBurst8En	<= '0';
			end if;
			-- Enable Write 4 Sector
			-- 4 sector remain transfer size
			-- 4 sector alignment address
			if ( rTrnLenCnt(47 downto 2)/=0 and 
				 rAXIAwAddr(10 downto 9)="00" ) then
				rBurst4En	<= '1';
			else
				rBurst4En	<= '0';
			end if;
			-- Enable Write 2 Sector
			-- 2 sector remain transfer size
			-- 2 sector alignment address
			if ( rTrnLenCnt(47 downto 1)/=0 and 
				 rAXIAwAddr(9)='0' ) then
				rBurst2En	<= '1';
			else
				rBurst2En	<= '0';
			end if;
		end if;
	End Process u_rBurstEn;
	
	-- Select BurstLen in sector unit
	-- Generate BurstLen at stChkWrRdy
	-- Using at stGenWrReq
	u_rBurstLen : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			-- Select burst size to transfer
			if ( rState=stChkWrRdy ) then
				-- Transfer 8 sector if
				-- At least 8 sector available in FIFO
				-- More than 8 sector space available in Buffer
				if ( rBurst8En='1' and AxiFfRdCnt(15 downto 6)/=0 and BuffWrCnt(30 downto 13)/=("11"&x"FFFF") ) then
					rBurstLen(3 downto 0)	<= x"8";
				-- Transfer 4 sector if
				-- At least 4 sector available in FIFO
				-- More than 4 sector space available in Buffer
				elsif ( rBurst4En='1' and AxiFfRdCnt(15 downto 5)/=0 and BuffWrCnt(30 downto 12)/=("111"&x"FFFF") ) then
					rBurstLen(3 downto 0)	<= x"4";
				-- Transfer 2 sector if
				-- At least 2 sector available in FIFO
				-- More than 2 sector space available in Buffer
				elsif ( rBurst2En='1' and AxiFfRdCnt(15 downto 4)/=0 and BuffWrCnt(30 downto 11)/=(x"FFFFF") ) then
					rBurstLen(3 downto 0)	<= x"2";
				-- Minimum burst size = 1 sector
				else
					rBurstLen(3 downto 0)	<= x"1";
				end if;
			else
				rBurstLen(3 downto 0)	<= rBurstLen(3 downto 0);
			end if;
		end if;
	End Process u_rBurstLen;
	
	-- Minimum AwLen is 1 sector => x"07"
	rAXIAwLen(2 downto 0)	<= "111";
	
	-- rBurstLen is valid in stGenWrReq at least 2 clock
	-- rBurstCnt load value of rAXIAwLen in stGenWrReq
	-- then rAXIAwLen can assign to rBurstCnt in stGenWrReq
	u_rAXIAwLen : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			-- unused rAXIAwLen(6)
			rAXIAwLen(6 downto 3)	<= rBurstLen(3 downto 0) - 1;
		end if;
	End Process u_rAXIAwLen;
	
	-- Axi write address valid
	-- Valid in stGenWrReq and valid until detect AXIAwReady='1'
	u_rAXIAwValid : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			if ( RstB='0' ) then
				rAXIAwValid     <= '0';
			else
				if ( rState=stGenWrReq ) then
					-- Command send completely
					if ( rAXIAwValid='1' and AXIAwReady='1' ) then
						rAXIAwValid     <= '0';
					else
						rAXIAwValid     <= '1';
					end if;
				else
					rAXIAwValid     <= '0';
				end if;
			end if;
		end if;
	End Process u_rAXIAwValid;
	
	-- Count data reading from FIFO
	u_rBurstCnt : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			-- Load value before start transferring
			-- In stGenWrReq state, wAxiFfRdEn is asserted for 2 clk cycles.
			-- rBurstCnt does not count down in stGenWrReq.
			-- So, end value of rBurstCnt for wAxiFfRdEn deasserted is 2.
			if ( rState=stGenWrReq ) then
				rBurstCnt(5 downto 0)	<= rAXIAwLen(5 downto 0);
			-- Decrement when read data from fifo
			elsif ( wAxiFfRdEn='1' ) then
				rBurstCnt(5 downto 0)	<= rBurstCnt(5 downto 0) - 1;
			else
				rBurstCnt(5 downto 0)	<= rBurstCnt(5 downto 0);
			end if;
		end if;
	End Process u_rBurstCnt;
	
	-- Generate Last of data in Axi Packet
	-- AXIwData is transferred from rAxiFfRdDataLat 
	-- which is latch register of read data from FIFO.
	-- So, rAXIwLast is designed by using 2-bit.
	-- Bit[0] is synchronus to read data from FIFO to deassert wAxiFfRdEn.
	-- Bit[1] is synchronous to rAxiFfRdDataLat for sending to AXI bus.
	u_rAXIwLast : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			-- Clear value before start transferring
			if ( rState=stChkWrRdy ) then
				rAXIwLast(1 downto 0)	<= "00";
			-- Last data in burst
			elsif ( rState=stWrTrans and AXIwReady='1' ) then
				rAXIwLast(1)	<= rAXIwLast(0);
				-- Last data has read from FIFO
				if ( rBurstCnt=2 ) then
					rAXIwLast(0)	<= '1';
				else
					rAXIwLast(0)	<= '0';
				end if;
			else
				rAXIwLast(1 downto 0)	<= rAXIwLast(1 downto 0);
			end if;
		end if;
	End Process u_rAXIwLast;

	-- Generate AXIwValid, this signal always Valid until end of transfer
	u_rAXIwValid : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			if ( RstB='0' ) then
				rAXIwValid	<= '0';
			else
				-- Transfer data 
				if ( rState=stWrTrans ) then
					-- End of transfer
					if ( rAXIwLast(1)='1' and AXIwReady='1' ) then
						rAXIwValid	<= '0';
					else
						rAXIwValid	<= '1';
					end if;
				else
					rAXIwValid	<= '0';
				end if;
			end if;
		end if;
	End Process u_rAXIwValid;
	
	-- Latch AxiFfRdData for ensure to met timing constrain
	u_rAxiFfRdDataLat : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			-- Latch data when first prepare data is coming
			if ( (rState=stGenWrReq and r2beatRdEn(1)='0') or 
			-- Latch data when Axi read before data already
				 (rState=stWrTrans and AXIwReady='1' and rAXIwValid='1')) then
				rAxiFfRdDataLat(511 downto 0)	<= AxiFfRdData(511 downto 0);
			else
				rAxiFfRdDataLat(511 downto 0)	<= rAxiFfRdDataLat(511 downto 0);
			end if;
		end if;
	End Process u_rAxiFfRdDataLat;
	
	-- Use this signal to allow wAxiFfRdEn asserting in stGenWrReq only 2 clock cycles.
	-- Also, it uses to enable rAxiFfRdDataLat to latch 1st read data from FIFO in stGenWrReq.
	u_r2beatRdEn : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			if ( RstB='0' ) then
				r2beatRdEn	<= "00";
			else
				-- when stGenWrReq r2beatRdEn is goto "01", "11" and hold until stChkWrRdy
				if ( rState=stGenWrReq ) then
					r2beatRdEn	<= r2beatRdEn(0) & '1';
				else
					r2beatRdEn	<= "00";
				end if;
			end if;
		end if;
	End Process u_r2beatRdEn;
	
								-- Read 2 data in state GenWrReq
								-- r2beatRdEn(1)='1' after state=stGenWrReq in 3rd clock cycle
								-- 1st data is latched to rAxiFfRdDataLat
								-- 2nd data is holded at output of FIFO
	wAxiFfRdEn		<= '1' when ((rState=stGenWrReq and r2beatRdEn(1)='0') or
								-- rAXIwLast is used to check that all data in current burst are read from FIFO.
								-- AXIwReady and rAXIwValid are used to confirm that current data has been received.
								-- Note: rAXIwValid='1' when rState=stWrTrans only
								 (rAXIwLast="00" and AXIwReady='1' and rAXIwValid='1'))
								else '0';
								
	-- AXI write response valid
    u_rAXIBValid : Process (Clk) Is
    Begin
        if ( rising_edge(Clk) ) then
            if ( RstB='0' ) then
                rAXIBValid      <= '0';
            else
                -- Set flag when current response
                if ( AXIBValid='1' ) then
                    rAXIBValid      <= '1';
                -- Latch until complete write
                elsif ( rState=stWtResp ) then
                    rAXIBValid      <= '0';
                else
                    rAXIBValid      <= rAXIBValid;
                end if;
            end if;
        end if;
    End Process u_rAXIBValid;
	
	-- Current DDR write byte pointer
	-- This signal is used to calculate BuffWrCnt for handshake with AxiMtRd.
	-- Pointer must be updated after data has been transferred by monitoring AXIBValid.
	u_rCurBuffWrAddr : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			-- Reset when start new command
			if ( rState=stIdle and AxiWrReq='1' ) then
				rCurBuffWrAddr(30 downto 9)	<= (others=>'0');
			-- End of write burst
			elsif ( rState=stWtResp and rAXIBValid='1' ) then
				rCurBuffWrAddr(30 downto 9)	<= rCurBuffWrAddr(30 downto 9) + ("00"&x"0000"&rBurstLen(3 downto 0));
			else
				rCurBuffWrAddr(30 downto 9)	<= rCurBuffWrAddr(30 downto 9);
			end if;
		end if;
	End Process u_rCurBuffWrAddr;
	
	-- Pulse to update CurBuffWrAddr value
	u_rCurBuffWrAdValid : Process (Clk) Is
	Begin
		if ( rising_edge(Clk) ) then
			-- Start new command : Set initial value
			if ( (rState=stIdle and AxiWrReq='1') or 
			-- End of write burst : Address is increased
				 (rState=stWtResp and rAXIBValid='1') ) then
				rCurBuffWrAdValid	<= '1';
			else
				rCurBuffWrAdValid	<= '0';
			end if;
		end if;
	End Process u_rCurBuffWrAdValid;
	
End Architecture rtl;
